from rdfcanon.identifier_issuer import IdentifierIssuer


class NDegreeResult:

    def __init__(self, hash: str, issuer: IdentifierIssuer):
        self.hash = hash
        self.issuer = issuer

    def __lt__(self, other: "NDegreeResult") -> bool:
        return self.hash < other.hash

    def __eq__(self, other: "NDegreeResult") -> bool:
        return self.hash == other.hash

    def __str__(self):
        return f"NDegreeResult(hash={self.hash}, issuer={self.issuer.prefix})"

    def __repr__(self):
        return self.__str__()



from rdflib import BNode, Dataset
from typing import MutableMapping
from rdflib import XSD
from rdflib.term import _toPythonMapping


class IdMap(MutableMapping[str, BNode]):
    def __init__(self, dct=None):
        self.dct = {} if dct is None else dct

    def __getitem__(self, key: str) -> BNode:
        return self.dct.setdefault(key, BNode(key))

    def __setitem__(self, key: str, value: BNode):
        self.dct[key] = value

    def __delitem__(self, key: str):
        return self.dct.__delitem__(key)

    def __iter__(self):
        return iter(self.dct)

    def __len__(self) -> int:
        return len(self.dct)


def parse_nquads_preserve_bnodes(file_path):

    for dt in (
        XSD.dateTime,
        XSD.boolean,
        XSD.double,
        XSD.decimal,
        XSD.integer,
        XSD.int,
        XSD.float,
    ):
        _toPythonMapping.pop(dt, None)

    skolem_graph = Dataset().parse(
            location=file_path, format="nquads", bnode_context=IdMap()
        )

    return skolem_graph



class IdentifierIssuer:
    def __init__(self, prefix):
        self.prefix = prefix
        self.counter = 0
        self.existing = {}

    def get_id(self, id: str) -> str:
        if id in self.existing:
            return self.existing[id]
        else:
            new_id = f"{self.prefix}{self.counter}"
            self.existing[id] = new_id
            self.counter += 1
            return new_id

    def hasId(self, id: str) -> bool:
        return id in self.existing

    def assign(self, other: "IdentifierIssuer"):
        for k in self.existing.keys():
            other.get_id(k)

    def copy(self) -> "IdentifierIssuer":
        new_issuer = IdentifierIssuer(self.prefix)
        new_issuer.counter = self.counter
        new_issuer.existing = self.existing.copy()
        return new_issuer


from enum import Enum


class RDFCanonTestCase:

    class Type(Enum):
        RDFC10MapTest = "rdfc:RDFC10MapTest"
        RDFC10EvalTest = "rdfc:RDFC10EvalTest"
        RDFC10NegativeEvalTest = "rdfc:RDFC10NegativeEvalTest"

    def __init__(self):
        self.id = ""
        self.type = RDFCanonTestCase.Type.RDFC10MapTest
        self.name = ""
        self.comment = ""
        self.input = ""
        self.expected = ""
        self.hash_algorithm = "sha256"


def getTestCaseOf(case: dict) -> RDFCanonTestCase:
    test_case = RDFCanonTestCase()
    test_case.id = case["id"]
    test_case.type = RDFCanonTestCase.Type(case["type"])
    test_case.name = case["name"]
    test_case.comment = case.get("comment", "")
    test_case.input = case["action"]
    test_case.expected = case.get("result", "")

    hash_algorithm = case.get("hashAlgorithm")
    if hash_algorithm == "SHA384":
        test_case.hash_algorithm = "sha384"
    else:
        test_case.hash_algorithm = "sha256"

    return test_case


def print_test_case(test_case: RDFCanonTestCase):
    print(f"Test Case ID: {test_case.id}")
    print(f"Type: {test_case.type}")
    print(f"Name: {test_case.name}")
    print(f"Comment: {test_case.comment}")
    print(f"Input: {test_case.input}")
    print(f"Expected: {test_case.expected}")
    print(f"Hash Algorithm: {test_case.hash_algorithm}")



import time


class RDFCanonTimeTicker:
    def __init__(self, max_time: int):
        self.current_time = 0

        if max_time < 0:
            raise ValueError("max_time must be non-negative")

        self.max_time = max_time

    def tick(self):
        current_time_ms = time.time_ns() / 1000000

        if self.current_time == 0:
            self.current_time = current_time_ms
            return

        time_passed = current_time_ms - self.current_time

        if time_passed >= self.max_time:
            raise TimeoutError("Operation timed out")



import json
from unittest.result import failfast
from rdfcanon.main import RDFCanon
from rdfcanon.nquads_custom_parser import parse_nquads_preserve_bnodes
from rdfcanon.rdfcanon_time_ticker import RDFCanonTimeTicker
from test.rdfcanon_test_case import RDFCanonTestCase, getTestCaseOf
import pytest


def read_manifest(verbose: bool = False) -> list[RDFCanonTestCase]:
    test_cases = []
    with open("test/manifest.jsonld", "r", encoding="utf-8") as f:
        manifest = json.load(f)

        for case in manifest["entries"]:
            test_case = getTestCaseOf(case)

            if verbose:
                print(f"Loaded test case: {test_case.id} - {test_case.name}")

            test_cases.append(test_case)

    print(f"Total test cases loaded: {len(test_cases)}")

    return test_cases


@pytest.mark.parametrize("test_case", read_manifest())
def test_one_case(test_case: RDFCanonTestCase):

    try:
        print("Parsing", "rdfcanon/test/" + test_case.input)

        dataset = parse_nquads_preserve_bnodes("test/" + test_case.input)

        canon: RDFCanon = RDFCanon(
            hash_algorithm=test_case.hash_algorithm,
            dataset=dataset,
            ticker=RDFCanonTimeTicker(3000),
        )

        result = canon.canonize()

        print("Canonization result:\n", result)

        assert result is not None, "Canonization result should not be None"

        with open("test/" + test_case.expected, "r", encoding="utf-8") as f:
            expected = f.read()

        if test_case.type == RDFCanonTestCase.Type.RDFC10EvalTest:
            assert_eval(test_case, expected, result)
        elif test_case.type == RDFCanonTestCase.Type.RDFC10MapTest:
            assert_map(test_case, json.loads(expected), canon.canon_issuer.existing)
        elif test_case.type == RDFCanonTestCase.Type.RDFC10NegativeEvalTest:
            failfast(f"Test case {test_case.id} was expected to fail but passed.")

    except Exception as e:
        if test_case.type != RDFCanonTestCase.Type.RDFC10NegativeEvalTest:
            raise e

    print(f"Test case {test_case.id} passed.")


def assert_eval(test_case: RDFCanonTestCase, expected: str, result: str):
    assert (
        expected == result
    ), f"Test case {test_case.id} failed: expected \n|{expected}|\n, got \n|{result}|\n"


def assert_map(test_case: RDFCanonTestCase, expected: dict, result: dict):
    match = False

    match = len(expected) == len(result)

    clean_result = {k: v[2:] for k, v in result.items() if v.startswith("_:")}

    if match:
        for k, v in expected.items():
            match = (k in clean_result) and (clean_result[k] == v)

            if not match:
                break

    assert (
        match
    ), f"Test case {test_case.id} failed: expected mapping\n|{expected.items()}|\ndoes not match result mapping\n|{clean_result.items()}|\n"



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="rdfcanon",
    version="0.1.0",
    packages=find_packages(),
    install_requires=[
        "rdflib==7.5.0",
        "sortedcontainers==2.4.0",
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="YoucTagh",
)



import hashlib


class HashWrapper:
    def __init__(self, algo="sha256"):
        self.algo = algo
        self.h = hashlib.new(algo)

    def update(self, data):
        self.h.update(data)

    def digest(self):
        return self.h.digest()

    def hexdigest(self):
        return self.h.hexdigest()

    def reset(self):
        self.h = hashlib.new(self.algo)



import itertools
import re
from rdflib import Dataset, URIRef
from rdflib.term import Node, BNode
from sortedcontainers import SortedDict
from rdfcanon.hash_wrapper import HashWrapper
from rdfcanon.identifier_issuer import IdentifierIssuer
from rdfcanon.n_degree_result import NDegreeResult
from rdfcanon.rdfcanon_time_ticker import RDFCanonTimeTicker


class RDFCanon:

    BLANK_A = "_:a"
    BLANK_Z = "_:z"

    def __init__(
        self,
        hash_algorithm: str,
        dataset: Dataset,
        ticker: RDFCanonTimeTicker = None,
    ):
        self.blank_id_to_quad_map: dict[str, list[tuple[Node, Node, Node, Node]]] = (
            dict()
        )
        self.blank_id_to_node_map = dict()
        self.non_normalized_blank_ids: set[str] = set()
        self.blank_id_to_normalized_blank_ids_map: dict[str, str] = dict()
        self.hash_to_blank_id_map: SortedDict[str, set[str]] = SortedDict()
        self.digest = HashWrapper(hash_algorithm)
        self.canon_issuer = IdentifierIssuer("_:c14n")
        self.canon_quads: list[str] = []
        self.dataset = dataset
        self.quads = dataset.quads()
        self.ticker = ticker

    def init_blank_id_quad_map(self, graph: Dataset):
        for quad in graph.quads():
            for node in quad:
                if isinstance(node, BNode):
                    blank_id = str(node)
                    if blank_id not in self.blank_id_to_quad_map:
                        self.blank_id_to_quad_map[blank_id] = []
                    self.blank_id_to_quad_map[blank_id].append(quad)

                    if blank_id not in self.blank_id_to_normalized_blank_ids_map:
                        self.blank_id_to_normalized_blank_ids_map[blank_id] = ""

    def init_non_normalized_blank_ids(self):
        self.non_normalized_blank_ids = set(self.blank_id_to_quad_map.keys())

    def prepare_quads_for_hashing(
        self, quad: tuple[Node, Node, Node, Node], blank_id: str
    ) -> str:

        subject = f"<{quad[0]}>"
        if isinstance(quad[0], BNode):
            subject = self.BLANK_A if str(quad[0]) == blank_id else self.BLANK_Z

        object = f"<{quad[2]}>" if isinstance(quad[2], URIRef) else quad[2].n3()
        if isinstance(quad[2], BNode):
            object = self.BLANK_A if str(quad[2]) == blank_id else self.BLANK_Z

        graph = f"<{quad[3]}>" if quad[3] != self.dataset.default_context.identifier else ".\n"
        if graph != "" and isinstance(quad[3], BNode):
            graph = self.BLANK_A if str(quad[3]) == blank_id else self.BLANK_Z
            graph = f"{graph} .\n"

        return f"{subject} <{quad[1]}> {object} {graph}"

    def hash_first_degree(self, blank_id: str) -> str:

        related = self.blank_id_to_quad_map[blank_id]
        prepared_quads: list[str] = []

        for quad in related:
            self.ticker.tick()
            prepared_quad = self.prepare_quads_for_hashing(quad, blank_id)
            prepared_quads.append(prepared_quad)

        prepared_quads.sort()

        self.digest.reset()

        for pq in prepared_quads:
            self.digest.update(pq.encode("utf-8"))

        return self.digest.hexdigest()

    def issueSimpleIds(self):
        simple: bool = True

        while simple:
            self.ticker.tick()
            simple = False
            self.hash_to_blank_id_map.clear()
            for blank_id in self.non_normalized_blank_ids:
                hash: str = self.hash_first_degree(blank_id)
                if hash not in self.hash_to_blank_id_map:
                    self.hash_to_blank_id_map[hash] = set()
                self.hash_to_blank_id_map[hash].add(blank_id)

        for hash in list(self.hash_to_blank_id_map.keys()):
            self.ticker.tick()
            blank_ids = self.hash_to_blank_id_map[hash]
            if len(blank_ids) == 1:
                blank_id = next(iter(blank_ids))
                self.canon_issuer.get_id(blank_id)
                self.non_normalized_blank_ids.remove(blank_id)
                del self.hash_to_blank_id_map[hash]
                simple = True
            else:
                print(
                    f"Hash collision for hash {hash} with blank IDs: {', '.join(blank_ids)}"
                )

    def issue_n_degree_ids(self):
        for k, v in self.hash_to_blank_id_map.items():

            hash_path_list: list[NDegreeResult] = []

            for blank_id in v:
                self.ticker.tick()
                if self.canon_issuer.hasId(blank_id):
                    continue

                blank_issuer = IdentifierIssuer("_:b")
                blank_issuer.get_id(blank_id)

                path: NDegreeResult = self.hash_n_degree_quads(blank_id, blank_issuer)
                hash_path_list.append(path)

            hash_path_list.sort()

            for result in hash_path_list:
                self.ticker.tick()
                result.issuer.assign(self.canon_issuer)

    def hash_n_degree_quads(self, id: str, issuer: IdentifierIssuer) -> NDegreeResult:
        return HashNDegreeQuads(self).hash(id, issuer)

    def make_canon_quads(self):

        for blank_id in self.blank_id_to_normalized_blank_ids_map.keys():
            canon_id = self.canon_issuer.get_id(blank_id)
            self.blank_id_to_normalized_blank_ids_map[blank_id] = canon_id

        sorted_quads: list[tuple[Node, Node, Node, Node]] = list(self.quads)
        lines: list[str] = []

        for quad in sorted_quads:
            subject = quad[0]
            if isinstance(subject, BNode):
                subject = BNode(
                    self.blank_id_to_normalized_blank_ids_map[str(quad[0])][2:]
                )
            object = quad[2]
            if isinstance(object, BNode):
                object = BNode(
                    self.blank_id_to_normalized_blank_ids_map[str(quad[2])][2:]
                )

            graph = quad[3]
            if isinstance(graph, BNode):
                graph = BNode(
                    self.blank_id_to_normalized_blank_ids_map[str(quad[3])][2:]
                )

            d = Dataset()
            if graph == self.dataset.default_context.identifier:
                d.add((subject, quad[1], object))
            else:
                d.add((subject, quad[1], object, graph))

            lines.append(d.serialize(format="nquads"))

        output = []
        for line in lines:
            line = self.canonicalize_nquads_escapes(line)
            line = ''.join(line.rsplit('\\n\\n', 1))
            line = ' .'.join(line.rsplit('  .', 1))
            output.append(line)

        output.sort()
        # print("\n".join(output))
        self.canon_quads = output

    def canonicalize_nquads_escapes(self, nq: str) -> str:
        def repl(match):
            ch = match.group(0)
            code = ord(ch)

            short = {
                0x08: r"\b",
                0x09: r"\t",
                0x0A: r"\n",
                0x0C: r"\f",
                0x0D: r"\r",
            }

            if code in short:
                return short[code]
            return f"\\u{code:04X}"

        # C0 controls + DEL
        return re.sub(r"[\x00-\x1F\x7F]", repl, nq)

    def canonize(self) -> str:

        self.ticker.tick()

        self.init_blank_id_quad_map(self.dataset)
        self.init_non_normalized_blank_ids()
        self.issueSimpleIds()
        self.issue_n_degree_ids()
        self.make_canon_quads()

        output = "\n".join(self.canon_quads) + "\n"

        return output if output != "\n" else ""


class HashNDegreeQuads:
    def __init__(self, outer: RDFCanon):
        self.data_to_hash: list[str] = []
        self.chosen_issuer = None
        self.chosen_path: list[str] = []
        self.outer = outer

    def append_to_path(
        self,
        related: str,
        path: list[str],
        issuer_copy: IdentifierIssuer,
        recursion_list: list[str],
    ):
        if self.outer.canon_issuer.hasId(related):
            path.append(self.outer.canon_issuer.get_id(related))
        else:
            if not issuer_copy.hasId(related):
                recursion_list.append(related)

            path.append(issuer_copy.get_id(related))

    def create_hash_to_related(self, id: str, issuer: IdentifierIssuer) -> SortedDict:
        hash_to_related: SortedDict[str, set[str]] = SortedDict()

        refer = self.outer.blank_id_to_quad_map[id]

        for quad in refer:
            self.outer.ticker.tick()
            for position in [0, 2, 3]:
                node = quad[position]
                if isinstance(node, BNode):
                    related = str(node)
                    if related != id:
                        hash: str = self.hash_related_blank_node(
                            related, quad, issuer, position
                        )
                        hash_to_related.setdefault(hash, set()).add(related)

        return hash_to_related

    def do_permutation(self, permutation: list[str], issuer: IdentifierIssuer):

        self.outer.ticker.tick()
        issuer_copy = issuer.copy()
        path: list[str] = []
        recursion_list: list[str] = []

        for related in permutation:
            self.outer.ticker.tick()
            self.append_to_path(related, path, issuer_copy, recursion_list)

            if len(self.chosen_path) > 0 and "".join(path) > "".join(self.chosen_path):
                return

        for related in recursion_list:
            self.outer.ticker.tick()
            result: NDegreeResult = self.outer.hash_n_degree_quads(related, issuer_copy)

            path.extend([issuer_copy.get_id(related), "<", result.hash, ">"])
            issuer_copy = result.issuer

            if len(self.chosen_path) > 0 and "".join(path) > "".join(self.chosen_path):
                return

        if len(self.chosen_path) == 0 or "".join(path) < "".join(self.chosen_path):
            self.chosen_path = path
            self.chosen_issuer = issuer_copy

    def hash(self, id: str, default_issuer: IdentifierIssuer) -> NDegreeResult:

        hash_to_related: SortedDict = self.create_hash_to_related(id, default_issuer)

        for k, v in hash_to_related.items():
            self.data_to_hash.append(k)
            self.chosen_path = []
            self.chosen_issuer = None

            permutations = [list(p) for p in itertools.permutations(v)]

            for permutation in permutations:
                self.outer.ticker.tick()
                self.do_permutation(permutation, default_issuer)

            self.data_to_hash.extend(self.chosen_path)
            default_issuer = self.chosen_issuer

        self.outer.digest.reset()
        self.outer.digest.update("".join(self.data_to_hash).encode("utf-8"))
        hash: str = self.outer.digest.hexdigest()
        return NDegreeResult(hash, default_issuer)

    def hash_related_blank_node(
        self,
        related: str,
        quad: tuple[Node, Node, Node, Node],
        issuer: IdentifierIssuer,
        position: int,
    ) -> str:

        id: str = None

        if self.outer.canon_issuer.hasId(related):
            id = self.outer.canon_issuer.get_id(related)
        elif issuer.hasId(related):
            id = issuer.get_id(related)
        else:
            id = self.outer.hash_first_degree(related)

        self.outer.digest.reset()
        self.outer.digest.update(self.get_position_tag(position).encode("utf-8"))

        if position != 3:
            self.outer.digest.update("<".encode("utf-8"))
            self.outer.digest.update(quad[1].encode("utf-8"))
            self.outer.digest.update(">".encode("utf-8"))

        self.outer.digest.update(id.encode("utf-8"))

        return self.outer.digest.hexdigest()

    def get_position_tag(self, position: int) -> str:
        if position == 0:
            return "s"
        elif position == 2:
            return "o"
        elif position == 3:
            return "g"
        else:
            raise Exception("Invalid position for hashing related blank node")
